#!/usr/bin/python
# -*- coding: utf-8 -*-

"""Convenience functions to manipulate an F-Spot database.

F-Spot, a photo manager, keeps it's meta-data in an SQLite3 database
(by default in ~/.gnome2/f-spot/photos.db for older versions of
F-Spot, and ~/.config/f-spot/photos.db for newer versions).  This
module provides a class offering easy access to some often needed (by
me) manipulations of the data in that database.
"""

__author__ = "Sven Utcke <sven.utcke@svenutcke.de"
__source__ = "http://www.svenutcke.de/sw/f-spot/cleanup/>"
__version__ = "$Revision: 1.7 $"
__license__ = """Copyright (C) 2011 Sven Utcke <sven.utcke@svenutcke.de>.

A nonexclusive license is granted to use this file for private
purposes.  A seperate license is required for comercial use, the
creation of copies or derivative work.

There is NO WARRANTY, to the extent permitted by law.
"""

import os
import time
import sqlite3
import shutil
import sys

import ckyorn
import my_uri

class FSpot():
    """Convenience functions to manipulate an F-Spot database.
    
    F-Spot, a photo manager, keeps it's meta-data in an SQLite3 database
    (by default in ~/.gnome2/f-spot/photos.db for older versions of
    F-Spot, and ~/.config/f-spot/photos.db for newer versions).  This
    module provides a class offering easy access to some often needed (by
    me) manipulations of the data in that database.
    """

    conn = None
    
    def check_db_version(self, backup):
    	"""Is this a known f-spot database version ?
    	
    	Figure out if this is a version we know about.  Currently all we know
    	about is db version 14 and 18 though...  Will exit if an unknown 
    	version is encountered."""
    	try:
    	    self.cursor.execute('''SELECT data FROM meta WHERE name="F-Spot Database Version"''')
    	except sqlite3.OperationalError, e:
    	    print e
    	except sqlite3.DatabaseError, e:
    	    print e
    	row = self.cursor.fetchone()
    	if (row == None):
            helptext = """I could not find any version information in
              """+self.name+""". This is not a good sign!  If you are
              really, really sure that this is an F-Spot database
              file, and if you are really, really adventurous, you
              could press 'y' now and see what happens"""
            if backup:
                helptext += """ (keeping in mind that a backup of your
                  database will be created in """+self.backup+")"
            helptext += """.  In that case, please run\n\n
              sqlite3 """+self.name+""" '.schema' && sqlite3 """+self.name+""" 'SELECT * FROM meta;'
              \n\nand send the output to bug@svenutcke.de
              \n\nAnswer 'n' if unsure.\n"""
    	    if ckyorn.ckyorn("Are you sure "+self.name+" is an F-Spot database?",
    	              helptext=helptext) == 'N':
    	        sys.exit(0)
    	    else:
    	        return None
    	elif (float(row[0]) != 14.0) and (float(row[0]) != 17.1) and (float(row[0]) != 18.0):
            helptext = """The F-Spot database format has been subject to
    	      frequent (and not necessarily compatible) changes.  Even
    	      worse, these different formats are generally undocumented,
    	      making them hard to support.  So what are your options?
              \n\nBefore you do anything, please run
              \n\n sqlite3 """+self.name+" '.schema' && sqlite3 "+self.name + """ 'SELECT * FROM meta;'
              \n\nand send the output to bug@svenutcke.de (unless
    	      you have already done so).  Once that is out of the way,
    	      what could you do then?
              \n\nYou could press 'y': Many of the changes do not affect
    	      us at all, so this script might work anyway.  Even if
    	      it does not, chances are it will simply fail without
    	      corrupting anything, or, if it does not fail outright, leave 
    	      the database in an only slightly but harmlessly inconsistent  
    	      state"""
            if backup:
                helptext += """. And if it should really corrupt your database, there's
                  the backup in """+self.backup+""" which you can simply copy back to
                  restore everything to working order"""
            helptext += """ (but make sure to check
              F-Spot right after executing this script, so as to catch
              any problems early on).
    	      \n\nNOTE: I _believe_ this ought to work for at least all versions
    	      between 7 and 18, and should fail otherwise.  However, only
    	      14, 17.1 and 18 were tested so far.
              \n\nYou could press 'n' or 'q': There is, after all, the
    	      slightly increased chance of database corruption --- so why
    	      not just send me the output of the commands above and let me
    	      adapt the script first?  It is the safer way, you know...\n\n"""
    	    if ckyorn.ckyorn("Unknown database Version "+row[0]+", proceed anyway?\n"
    	              "Press '?' if unsure.", 
    	              helptext=helptext) == 'N':
    	        sys.exit(0)
    	    else:
    	        return row[0]
    	else:                       # known format
    	    return row[0]    

    def Images(self, tag=None, tagged=True):
        """Iterate over all (tagged) images (returning the decoded and encoded path)
        
        Only iterate over the images tagged as 'tag', if specified.
        By default, return all tagged images.  This needs to iterate
        both over paths stored in the Photos table, as well as paths
        stored in the Photo_versions table, taking into account that
        at least DB version 18 stores the same path in both.  Sigh.
        """ 
        # we need two cursors, one for each table
        photos_cursor = self.conn.cursor()
        versions_cursor = self.conn.cursor()
        # TODO: parametrise this some more!
        if not tag:
            if tagged:
                # select all tagged images
                photos_cursor.execute('''SELECT decoded, encoded FROM photos, photo_tags, urimap 
                                         WHERE photos.id=photo_tags.photo_id 
                                         AND '''+self.uri('photos')+'''=encoded
                                         GROUP BY photos.id
                                         ORDER BY '''+self.uri('photos'))
                versions_cursor.execute('''SELECT decoded, encoded FROM photo_versions, photo_tags, urimap
                                         WHERE version_id<>1
                                         AND photo_versions.photo_id=photo_tags.photo_id 
                                         AND '''+self.uri('photo_versions')+'''=encoded
                                         GROUP BY '''+self.uri()+'''
                                         ORDER BY '''+self.uri('photo_versions'))
            else:
                # select _all_ images
                photos_cursor.execute('''SELECT decoded, encoded FROM photos, urimap
                                         WHERE '''+self.uri()+'''=encoded 
                                         GROUP BY photos.id
                                         ORDER BY encoded''')
                versions_cursor.execute('''SELECT decoded, encoded FROM photo_versions, urimap
                                           WHERE version_id<>1
                                           AND '''+self.uri()+'''=encoded 
                                           GROUP BY '''+self.uri()+'''
                                           ORDER BY encoded''')
        else:
            # all images with tag tag
            photos_cursor.execute('''SELECT decoded, encoded FROM photos, photo_tags, tags, urimap
                                     WHERE tags.name=?
                                     AND tags.id=photo_tags.tag_id 
                                     AND photos.id=photo_tags.photo_id 
                                     AND '''+self.uri('photos')+'''=encoded
                                     GROUP BY photos.id
                                     ORDER BY encoded''', [tag,])
            versions_cursor.execute('''SELECT decoded, encoded FROM photo_versions, photo_tags, tags, urimap
                                     WHERE version_id<>1
                                     AND tags.name=?
                                     AND tags.id=photo_tags.tag_id
                                     AND photo_versions.photo_id=photo_tags.photo_id 
                                     AND '''+self.uri('photo_versions')+'''=encoded
                                     GROUP BY '''+self.uri()+'''
                                     ORDER BY encoded''', [tag,])
        for cursor in [photos_cursor, versions_cursor]:
            for row in cursor:
                decoded = row[0]
                encoded = row[1]
                yield decoded[7:], encoded[7:]

    def isVersion(self, path):
        """Does this file have multiple versions, and which one are we?

        Returns 'None' if the path doesn't exist in the database,
        returns 0 if the path exists in 'photos', and no version with 
                  'version_id>1' exists in 'photo_versions'.
        returns 1 if the path exists in 'photos', and at least one 
                  version with 'version_id>1' exists in 'photo_versions'.
        returns the 'version_id' if the path exists in 'photo_versions'.

        NOTE: there is a BUG in at least f-spot 0.8.0 which will not 
              always encode all encodeable characters, which is a mayor 
              pain, and will result in this function erroneously returning 
              'None'
        """
        if not path:
            return None
        # try to find this in photo_versions first.
        # returns the 'version_id' if the path exists in 'photo_versions'.
        self.cursor.execute('''SELECT a.version_id, b.version_id 
                               FROM photo_versions as a, photo_versions as b, urimap
                               WHERE urimap.decoded="file://"||?
                               AND '''+self.uri('a')+'''=urimap.encoded
                               AND a.photo_id=b.photo_id 
                               ORDER BY b.version_id DESC LIMIT 1''',
                            [path,])
        row = self.cursor.fetchone()
        if row:
            if (int(row[0]) > 1):
                return int(row[0])
        # Returns 'None' if the path doesn't exist in the database,
        # returns 0 if the path exists in 'photos', and no version with 
        #           'version_id>1' exists in 'photo_versions'.
        # returns 1 if the path exists in 'photos', and at least one 
        #           version with 'version_id>1' exists in 'photo_versions'.
        # TODO: do this in one query!
        self.cursor.execute('''SELECT id from photos, urimap
                               WHERE urimap.decoded="file://"||?
                               AND '''+self.uri()+'''=urimap.encoded''',
                            [path,])
        row = self.cursor.fetchone()
        if not row:
            return None
        id = int(row[0])
        self.cursor.execute('''SELECT EXISTS( SELECT 1 from photo_versions 
                               WHERE photo_id=? AND version_id>1)''',
                            [id,])
        row = self.cursor.fetchone()
        if row:
            return int(row[0])
        else:
            return -1       # can't happen

    def __urldecode(self, url):
        """return the decoded uri.

        This assumes that the uri passed to it was previously
        urlencoded
        (http://www.blooberry.com/indexdot/html/topics/urlencoding.htm).
        This _will_ give rise to problems if the uri passed to it
        looks like it was encoded but wasn't, or if it contains both
        encoded characters as well as un-encoded non-ASCII
        characters."""
        try:
            res =  my_uri.unquote(url)
        except UnicodeEncodeError, e:
            print type(e)
            print e
            res = url
        except Exception, e:
            print type(e)
            print e
            res = url
        return res
            
    def delete(self, path=None, index=None):
        '''Delete an entry from the DB by either path or index.

        Currently, only path is implemented, and even this causes some serious problems.'''
        # Easy: files without versions, and versions that are not the default version 
        # (although I feel that those ought to renumber version numbers, this is not 
        # wh\t f-spot itself does...
        # if the version to delete is the one in default_version_id, that gets set to 
        # the latest version
        
        # Also quite easy: file in photo_versions, with version_id>1
        # Here, we can simply delete the entry from photo_versions.
        # In addition, we need to set the default_version_id to the highest 
        # existing version_id (or 1) if we deleted the default version.
        if self.pretend or self.verbose:
            print "Deleting", path, "from DB"
            # version_id>1?
            self.cursor.execute('''SELECT '''+self.uri()+''', version_id, photo_id FROM photo_versions
                                     WHERE '''+self.uri()+''' IN 
                                    (SELECT urimap.encoded FROM photo_versions, urimap 
                                       WHERE urimap.decoded="file://"||? 
                                       AND urimap.encoded='''+self.uri('photo_versions')+'''
                                       AND version_id>1)''', [path,])
            id = -1
            version_id = -1
            for row in self.cursor:
                version_id = row[1]
                id = row[2]
                print row[0], "has version "+str(row[1])+", deleting from photo_versions"
            # default_version_id needs resetting?
            self.cursor.execute('''SELECT default_version_id FROM photos
                                     WHERE id=?''', [id,])
            for row in self.cursor:
                if row[0] == version_id:
                    print "Also need to reset default_version_id"
        if not self.pretend:
           self.cursor.execute('''DELETE FROM photo_versions
                                     WHERE '''+self.uri()+''' IN 
                                    (SELECT urimap.encoded FROM photo_versions, urimap 
                                       WHERE urimap.decoded="file://"||? 
                                       AND urimap.encoded='''+self.uri('photo_versions')+'''
                                       AND version_id>1)''', [path,])
           # fix default_version_id
           # This fixes exactly one version_id (for the lowest id), 
           # but usually only one should need fixing...
           self.cursor.execute('''UPDATE photos SET default_version_id=
                                   (SELECT version_id FROM photos, photo_versions 
                                     WHERE photo_id=id AND default_version_id NOT IN 
                                      (SELECT version_id FROM photo_versions 
                                        WHERE photo_id=id) GROUP BY id) 
                                   WHERE id=
                                   (SELECT id FROM photos, photo_versions 
                                     WHERE photo_id=id AND default_version_id NOT IN 
                                      (SELECT version_id FROM photo_versions 
                                        WHERE photo_id=id) GROUP BY id)''')
        # Easy case: the file is _only_ mentioned in photos:
        # TODO: also delete version_id=1 from photo_versions, if it exist!
        if self.pretend or self.verbose:
            # look up in photos
            self.cursor.execute('''SELECT '''+self.uri()+''' FROM photos 
                                     WHERE '''+self.uri()+''' IN 
                                    (SELECT urimap.encoded FROM photos, urimap 
                                       WHERE urimap.decoded="file://"||? 
                                       AND urimap.encoded='''+self.uri('photos')+'''
                                       AND NOT EXISTS( SELECT 1 from photo_versions 
                                         WHERE photos.id=photo_versions.photo_id 
                                         AND version_id>1))''', [path,])
            for row in self.cursor:
                print row[0][7:], "has no versions, deleting from photos"
            # look up in photo_versions
            self.cursor.execute('''SELECT '''+self.uri()+''', version_id FROM photo_versions
                                     WHERE '''+self.uri()+''' IN 
                                       (SELECT urimap.encoded FROM photo_versions AS p, urimap 
                                          WHERE urimap.decoded="file://"||? 
                                          AND urimap.encoded='''+self.uri('p')+'''
                                          AND NOT EXISTS( SELECT 1 from photo_versions 
                                            WHERE p.photo_id=photo_versions.photo_id 
                                            AND version_id>1))''', [path,])
            for row in self.cursor:
                print row[0][7:], "version", row[1], "has no further versions, deleting from photo_versions"
        if not self.pretend:
            # delete from photos
            self.cursor.execute('''DELETE FROM photos WHERE '''+self.uri()+''' IN 
                                    (SELECT urimap.encoded FROM photos, urimap 
                                       WHERE urimap.decoded="file://"||? 
                                       AND urimap.encoded='''+self.uri('photos')+'''
                                       AND NOT EXISTS( SELECT 1 from photo_versions 
                                         WHERE photos.id=photo_versions.photo_id 
                                         AND version_id>1))''', [path,])
            # also delete version 1 from photo_versions
            self.cursor.execute('''DELETE FROM photo_versions 
                                     WHERE '''+self.uri()+''' IN 
                                       (SELECT urimap.encoded FROM photo_versions AS p, urimap 
                                          WHERE urimap.decoded="file://"||? 
                                          AND urimap.encoded='''+self.uri('p')+'''
                                          AND NOT EXISTS( SELECT 1 from photo_versions 
                                            WHERE p.photo_id=photo_versions.photo_id 
                                            AND version_id>1))''', [path,])
        # so here's the difficult one: deleting the base version if 
        # there are other versions, in which case we would need to 
        # remove the base version from photo_versions (if it exists), 
        # make the next version the base version, cp that over to 
        # photos, and possible reset the default_version_id
        # look up in photo_versions
        self.cursor.execute('''SELECT '''+self.uri()+''', id FROM photos
                                 WHERE '''+self.uri()+''' IN 
                                   (SELECT urimap.encoded FROM photos AS p, urimap 
                                      WHERE urimap.decoded="file://"||? 
                                      AND urimap.encoded='''+self.uri('p')+'''
                                      AND EXISTS( SELECT 1 from photo_versions 
                                        WHERE p.id=photo_versions.photo_id 
                                        AND version_id>1))''', [path,])
        photo_id=-1
        for row in self.cursor:
            photo_id = row[1]
            if self.verbose or self.pretend:
                print row[0][7:], "is base version, deleting from photo_versions"
        if self.verbose or self.pretend:
            # find the next version up and rename that to 1
            self.cursor.execute('''SELECT version_id FROM photo_versions
                                     WHERE photo_id=? 
                                     AND version_id IN 
                                       (SELECT version_id FROM photo_versions 
                                          WHERE photo_id=? AND version_id>1 ORDER BY version_id LIMIT 1)''',
                                [photo_id, photo_id])
            version_id=-1
            for row in self.cursor:
                version_id = row[0]
                print "renaming version", version_id, "to 1"
                # copy that version over to photos
                print "copying the new version 1 to photos"
            # rename default_version_id if necessary
            self.cursor.execute('''SELECT default_version_id FROM photos
                                     WHERE id=?''', [photo_id,])
            for row in self.cursor:
                if row[0] == version_id:
                    print "Also need to reset default_version_id"
        if not self.pretend and not photo_id==-1:
            # delete base version from photo_versions --- 
            # of course this could be done using photo_id...
            self.cursor.execute('''DELETE FROM photo_versions
                                     WHERE '''+self.uri()+''' IN 
                                       (SELECT urimap.encoded FROM photo_versions AS p, urimap 
                                          WHERE urimap.decoded="file://"||? 
                                          AND urimap.encoded='''+self.uri('p')+'''
                                          AND version_id=1
                                          AND EXISTS( SELECT 1 from photo_versions 
                                            WHERE p.photo_id=photo_versions.photo_id 
                                            AND version_id>1))''', [path,])
            # make this version_id=1
            self.cursor.execute('''UPDATE photo_versions SET version_id=1 
                                     WHERE photo_id=? 
                                     AND version_id IN 
                                       (SELECT version_id FROM photo_versions 
                                          WHERE photo_id=? ORDER BY version_id LIMIT 1)''',
                                [photo_id, photo_id])
            # copy that version over to photos
            # TODO: use update instead of replace! BUT: can only update one field!
            self.cursor.execute('''REPLACE INTO photos (id, '''+self.uri_elements()+''', 
                                                        time, description, 
                                                        roll_id, default_version_id)
                                     SELECT photo_id, '''+self.uri_elements('photo_versions')+''',
                                            time, description, 
                                            roll_id, default_version_id 
                                       FROM photo_versions, photos 
                                       WHERE id=photo_id and photo_id=? and version_id=1''',
                                [photo_id,])
            # fix default_version_id
            self.cursor.execute('''UPDATE photos SET default_version_id=1 
                                    WHERE id=? AND default_version_id NOT IN 
                                      (SELECT version_id FROM photo_versions 
                                        WHERE photo_id=id)''', [photo_id,])


    def uri(self, prefix=None):
        if float(self.version) < 17:
            if prefix:
                return prefix+'.uri'
            else:
                return 'uri'
        elif float(self.version) >= 17:
            if prefix:
                return prefix+".base_uri||'/'||"+prefix+".filename"
            else:
                return "base_uri||'/'||filename"

    def uri_elements(self, prefix=None):
        if float(self.version) < 17:
            if prefix:
                return prefix+'.uri'
            else:
                return 'uri'
        elif float(self.version) >= 17:
            if prefix:
                return prefix+".base_uri, "+prefix+".filename"
            else:
                return "base_uri, filename"

    def __init__(self, databasename=None, backup=True, verbose=False, pretend=False):
        # * open the given database, or one of 
        #   ~/.config/f-spot/photos.db
        #   ~/.gnome2/f-spot/photos.db
        # * make a backup if backup is True,
        # * find out what version we are dealing with.
        # * create a temporary table linking (possibly) url-encoded names to their decoded counterparts
        self.verbose = verbose
        self.pretend = pretend
        if databasename:
            names = [databasename,]
        else:
            names = ['~/.config/f-spot/photos.db', '~/.gnome2/f-spot/photos.db']
        for name in names:
            self.name = os.path.abspath(os.path.expanduser(os.path.expandvars(name)))
            if backup:
                self.backup = self.name + time.strftime("-%Y%m%d-%H:%M:%S") # TODO: race condition!
            else:
                self.backup = None
            # connect to the database
            try:
                self.conn = sqlite3.connect(self.name)
            except sqlite3.OperationalError, e:
                continue
            self.cursor = self.conn.cursor()
            self.version = self.check_db_version(backup)
            # we only make a backup copy if this is a known version, and if backup was set:
            if backup:
                # make a backup of the db
                try:
                    shutil.copy2(self.name, self.backup)
                except OSError, e:
                    print e
                    print "Could not make a backup copy from", self.name, "to", self.backup
                    raise
            break
        if not self.conn:
            sys.exit(1)
        else:
            if self.verbose:
                print "opened file ", self.name
        # create a temporary table linking (possibly) url-encoded names to their decoded counterparts
        if self.verbose:
            print "Building table of decoded filenames.  ",
        self.conn.create_function("urldecode", 1, self.__urldecode)
        self.cursor.execute('''CREATE TEMPORARY TABLE urimap (encoded, decoded)''')
        # fill this according to version
        self.cursor.execute('''INSERT INTO urimap (encoded, decoded) 
                                 SELECT '''+self.uri()+''', urldecode('''+self.uri()+''') FROM photos WHERE '''+self.uri()+''' LIKE "file://%"''')
        self.cursor.execute('''INSERT INTO urimap (encoded, decoded) 
                                 SELECT '''+self.uri()+''', urldecode('''+self.uri()+''') FROM photo_versions WHERE '''+self.uri()+''' LIKE "file://%"''')
        # and index those
        self.cursor.execute('''CREATE INDEX idx_decoded ON urimap (decoded)''')
        self.cursor.execute('''CREATE INDEX idx_encoded ON urimap (encoded)''')
        if self.verbose:
            print "Done"        


            

            
    def close(self):
        """closes the database connection. Should be the last action called."""
        self.conn.commit()
        self.conn.close()

if __name__ == "__main__":
    ## This ought to fail
    #db = FSpot('/dev/random')
    #db = FSpot('/tmp/photos.db', pretend=True)
    db = FSpot('/tmp/f-spot/photos.db', pretend=True)
    #db = FSpot()
    i=0
    for x in db.Images(tagged=False):
        print x[0], x[1]
        i+=1
    print 'Number of files:', i
    i=0
    for x in db.Images():
        i+=1
    print 'Number of tagged files:', i
    i=0
    for x in db.Images(tag='Favoriten'):
        print x[0]
        i+=1
    print 'Number of "Favoriten":', i
    i=0
    for x in db.Images(tag='Original'):
        print x[0]
        i+=1
    print 'Number of "Original":', i
    print 'version of "IMG_2328.JPG.jpg"', db.isVersion('/home/utcke/2011/02/04/IMG_2328.JPG.jpg')
    print 'version of "IMG_2328.JPG-1 (Modified).jpg"', db.isVersion('/home/utcke/2011/02/04/IMG_2328.JPG-1 (Modified).jpg')
    print 'version of "IMG_5482.JPG"', db.isVersion('/media/disk/DCIM/112CANON/IMG_5482.JPG')
    print 'version of "curslack341.jpg"', db.isVersion('/home/utcke/www/Private/Haus/curslack341.jpg')
    print 'version of "/home/utcke/Photos/2010/02/07/img_0437.jpg"', db.isVersion('/home/utcke/Photos/2010/02/07/img_0437.jpg')
    print 'version of "img_0696 (Bearbeitet).jpg"', db.isVersion('/home/utcke/Photos/2011/05/01/img_0696 (Bearbeitet).jpg')
    print 'version of "Bildschirmfoto.png"', db.isVersion('/home/utcke/Bilder/Fotos/2011/03/28/Bildschirmfoto.png')
    print 'version of "ich.jpg"', db.isVersion('/home/utcke/Bilder/Fotos/2011/06/13/ich.jpg')
    print 'version of "ich (Bearbeitet (3)).jpg"', db.isVersion('/home/utcke/Bilder/Fotos/2011/06/13/ich (Bearbeitet (3)).jpg')
    #print 'version of "ich äöü.jpg"', db.isVersion('/home/utcke/Bilder/Fotos/2011/06/13/ich äöü.jpg')
    print 'version of "3128440683 (Bearbeitet (5)).jpg"', db.isVersion('/home/utcke/Fotos/2011/05/04/3128440683 (Bearbeitet (5)).jpg')

    db.cursor.execute('''SELECT * FROM urimap WHERE encoded <> decoded''')
    rows = db.cursor.fetchall()
    for row in rows:
        print row

    db.pretend = True
    db.verbose = True
    for decoded, encoded in db.Images(tagged=False):
        found = False
        if os.path.exists(decoded): found = True
        elif os.path.exists(encoded): found = True
        if not found:
            db.delete(decoded)

    print 'version of "3128440683.jpg"', db.isVersion('/home/utcke/Fotos/2011/05/04/3128440683.jpg')

    db.close()
