#!/usr/bin/python
# -*- coding: utf-8 -*-

"""Rescan F-Spot's database, removing any files no longer found on disk.

Rescan F-Spot's database (default at ~/.gnome2/f-spot/photos.db),
removing any files no longer found on disk.  Be careful in case your
images reside on a removeable disk :-)
"""

__author__ = "Sven Utcke <sven.utcke@svenutcke.de"
__source__ = "http://www.svenutcke.de/sw/f-spot/cleanup/>"
__version__ = "$Revision: 1.10 $"
__license__ = """Copyright (C) 2010, 2011 Sven Utcke <sven.utcke@svenutcke.de>.

A nonexclusive license is granted to use this file for private
purposes.  A seperate license is required for comercial use, the
creation of copies or derivative work.

There is NO WARRANTY, to the extent permitted by law.
"""

from optparse  import OptionParser
import shutil
import os.path
import time
import sqlite3
import sys
import hashlib
import glob
import urllib

# my own f-spot DB library
from f_spot import FSpot
from ckyorn import ckyorn

#=======================================================================
#
# add version output
#
def version(option, opt, value, pars):
    """Output version information and die."""
    import f_spot_cleanup_version
    versionString = "$Id: f_spot_cleanup.py,v 1.10 2011/08/01 17:02:39 utcke Exp utcke $".split()
    print "f_spot_cleanup", f_spot_cleanup_version.version, "(rcs " + versionString[2] + ")"
    print """Copyright (C) 2010,2011 Sven Utcke <sven.utcke@svenutcke.de>.

A nonexclusive license is granted to use this file for private
purposes.  A seperate license is required for comercial use, the
creation of copies or derivative work.

There is NO WARRANTY, to the extent permitted by law.

Written by Sven Utcke <sven.utcke@svenutcke.de>."""
    sys.exit(0)


#=======================================================================
#
# usage string, prepended to output on --help
#
usage = """

%prog  will scan  F-Spot's database,  either  the one  given with  the
--database   switch,   or   one  of   ~/.config/f-spot/photos.db   and
~/.gnome2/f-spot/photos.db  (in that  order), for  files no  longer on
disk and remove those from  the database.  By default, only references
to files  where the  file is missing,  but its parent  directory still
exists,  are removed  (this helps  to  avoid problems  in cases  where
photos  reside  on  a  volume  not currently  mounted).   Use  the  -f
(--force) option to override this behaviour, or the -i (--interactive)
option to confirm  the removal of each individual  entry.  A backup of
the database  file is  created first, under  the name of  the database
file with date and time appended.

In addition, all GNOME thumbnails (representing these files) in either
~/.thumbnails or the directory referenced by the --thumbnaildir option
are removed too.

Note:  This  currently  only  uses   the  URI's  md5sum  to  find  the
thumbnails, but  does not check the  thumbnail's Thumb::URI attribute,
so  that in very  rare cases  the wrong  thumbnail might  get deleted.
This  is,  however,  not  a  problem,  as  those  thumbnails  will  be
regenerated if needed.

Version 0.3  should now be able to  work with any of  Python 2.5, 2.6,
and 2.7, even where accented  characters are concerned; should be able
to remove references to files even if derived versions exist; and will
probably work  with at  least all database-versions  between 7  and 18
(tested with 14 and 18).

usage: %prog [ -fhinVv ] [ Options ] """

# Parse command line
parser = OptionParser(usage=usage)
parser.add_option("-V", "--version", dest="version", action="callback", 
                  callback=version,
                  help="Print version information and exit.")
parser.add_option("-f", "--force", action="store_true", dest="force", 
                  help="Also delete references to files in non-existing "
                  "directories.  Be careful, as this will cause problems "
                  "in cases where F-Spot references photos on currently "
                  "unmounted volumes.")
parser.add_option("-i", "--interactive", dest="interactive", 
                  action="store_true",
                  help="Prompt before each removal.")
parser.add_option("-n", "--dry-run", dest="pretend", action="store_true", 
                  default=False,
                  help="Print the commands that would be executed, "
                  "but do not execute them.")
parser.add_option("-v", "--verbose", dest="verbose", action="store_true",
                  help="Explain what is being done.")
parser.add_option("--database", dest="databasefile", action="store", 
                  type="string", default=None,
                  help="F-Spot's database (one of ~/.config/f-spot/photos.db "
                  "or ~/.gnome2/f-spot/photos.db by "
                  "default).  A copy will be made before any changes are "
                  "attempted.")
parser.add_option("--thumbnaildir", dest="thumbdir", action="store", 
                  type="string", default="$HOME/.thumbnails",
                  help="Gnome's thumbnail directory ( ~/.thumbnails by "
                  "default).")
(options, args) = parser.parse_args()

# expand path to thumbdir:
options.thumbdir = os.path.abspath(os.path.expanduser(os.path.expandvars(options.thumbdir)))

# open DB:
if options.databasefile:
    options.databasefile = os.path.abspath(os.path.expanduser(os.path.expandvars(options.databasefile)))

db = FSpot(options.databasefile, pretend=options.pretend, 
           verbose=options.verbose, backup=(not options.pretend))

olddir = ""
# find all referenced files
for decoded, encoded in db.Images(tagged=False):
    if options.verbose:
        print decoded, encoded
    found = False
    if os.path.exists(decoded): found = True
    elif os.path.exists(encoded): found = True
    delete = not found
    if not found:
        # find thumbnails
        thumbnails = []
        for path in [decoded, encoded]:
            m = hashlib.md5()       # TODO: surely there must be a better way?
            # this uses url-encoding
            m.update('file://'+urllib.pathname2url(path.encode('utf-8')))
            m.hexdigest()
            thumbnails += (glob.glob(options.thumbdir + '/*/' + m.hexdigest() + '.png'))
        # delete if --force
        if options.force:
            delete = True
        elif options.interactive:
            if not os.path.exists(os.path.dirname(decoded)) and not os.path.exists(os.path.dirname(encoded)):
                print "\nAttention: Neither path", os.path.dirname(decoded), 
                print "(nor", os.path.dirname(encoded), "does exist!"
            if ckyorn("Delete "+decoded+" from DB?", 
                      helptext="This will delete all references to "+decoded+
                      " from the database "+options.databasefile+
                      ".\nAnswer 'n' if unsure, in particular if you were "+
                      "warned about non-existing directories above!\n") == 'Y':
                delete = True
            else:
                delete = False
        elif os.path.exists(os.path.dirname(decoded)) or os.path.exists(os.path.dirname(encoded)):
            delete = True
        else:
            delete = False
            if options.verbose:
                print "Neither path", os.path.dirname(decoded), 
                print "(nor", os.path.dirname(encoded), ") does exist!",
                print "Ignoring missing file", decoded
        if delete:
            db.delete(decoded)
            for thumbnail in thumbnails:
                if options.verbose or options.pretend:
                    print "Removing thumbnail", thumbnail
                if not options.pretend:
                    try:
                        os.unlink(thumbnail)
                    except OSError:
                        pass
db.close()
